package OntoDLPlus;

import OntoDLPlus.*;

import javax.xml.crypto.Data;
import java.util.*;
import java.lang.StringBuilder;
import java.util.stream.Collectors;

public class Ontology {
    private String name;
    private HashSet<Concept> concepts;
    private HashSet<String> individuos;
    private HashSet<Relation> relacoes;
    private HashSet<Triple> triplos;
    private List<SemError> erros;
    private List<DataProperty> atributos;
    private HashMap<Triple,Restriction> restrictions;

    public Ontology() {
        this.name = "";
        this.concepts = new HashSet<Concept>();
        this.individuos = new HashSet<String>();
        this.relacoes = new HashSet<Relation>();
        this.triplos = new HashSet<Triple>();
        this.erros = new ArrayList<SemError>();
        this.atributos = new ArrayList<DataProperty>();
        this.restrictions = new HashMap<Triple,Restriction>();
    }

    public Ontology(final String n) {
        this.name = n;
        this.concepts = new HashSet<Concept>();
        this.individuos = new HashSet<String>();
        this.relacoes = new HashSet<Relation>();
        this.triplos = new HashSet<Triple>();
        this.erros = new ArrayList<SemError>();
        this.atributos = new ArrayList<DataProperty>();
        this.restrictions = new HashMap<Triple,Restriction>();
    }

    public void setName(String s) {this.name=s;}

    public String getName() {return this.name;}

    public void addRelation (String s){
        if(!this.containsRelation(s)) this.relacoes.add(new Relation(s));
        else this.addSemError("ERROR: Multiple definition of relation: "+s);
    }

    public void addRelation (String n, String d, String c, Set<String> p){
        if(!this.containsRelation(n)) {
            if((this.containsConcept(d) || d==null) && (this.containsConcept(c) || c==null)) {
                this.relacoes.add(new Relation(n, d, c, p));
            } else {this.addSemError("ERROR: Invalid domain/codomain for relation: "+n);}
        }
        else this.addSemError("ERROR: Multiple definition of relation: "+n);
    }

    public boolean containsIndividual(final String s) {
        return this.individuos.contains(s);
    }

    public void addIndividual (String s) {
        if (!this.containsIndividual(s)) this.individuos.add(s);
        else this.addSemError("ERROR: Multiple definition of individual: "+s);
    }

    public boolean containsConcept(final String s) {
        for (final Concept c : this.concepts) {
            if (c.getConcept().equals(s)) {
                return true;
            }
        }
        return false;
    }

    public void addRestriction (String l, String rl, String rgt, Restriction r) {
        Triple t= new Triple(l,rl,rgt,this);
        this.restrictions.put(t.clone(),r.clone());
    }


    public Restriction createRestriction (Triple t, boolean b) {
        Restriction res= new Restriction(t, b);
        return res;
    }

    public Restriction createRestriction (String left, String relation, String right) {
        Triple t = new Triple(left, relation, right);
        Restriction res= new Restriction(t, false);
        return res;
    }

    public Triple genTriple (String left, String relation, String right) {
        return new Triple(left,relation,right,this);
    }

    public void genDataProperty (String prop, String ind, String conc, String value) {
        for(Concept c: this.concepts) {
            if(c.getConcept().equals(conc)) {
                DataProperty d= c.getProperty(prop);
                if(d!=null) {
                    d.setIndName(ind);
                    try {
                        d.setValue(value);
                        this.atributos.add(d.clone());
                    } catch (Exception e){
                        StringBuilder s= new StringBuilder();
                        s.append(e.getMessage());
                        s.append(" Na propriedade ");
                        s.append(d.getPropName());
                        s.append(" do indivíduo ");
                        s.append(d.getIndName());
                        s.append(".\n");
                        this.addSemError(s.toString());
                    }
                }
            }
        }
    }

    public boolean isRelated (String f, String c) {
        boolean res = false;
        for(Triple t: this.triplos) {
            if(t.getLeft().equals(f)) {
                if(t.getRelation().equals("iof")||t.getRelation().equals("isa")){
                    if(t.getRight().equals(c)) {
                        res=true;
                    }
                }
            }
        }
        return res;
    }

    public boolean isRestricted (Triple tested, Triple test) {
        boolean res=true;
        if(!tested.getLeft().equals(test.getLeft())) {
            if(!isRelated(tested.getLeft(),test.getLeft()))
                res=false;
        }

        if(!tested.getRelation().equals(test.getRelation())) {
            res=false;
        }

        if(!tested.getRight().equals(test.getRight())) {
            if(!isRelated(tested.getRight(),test.getRight()))
                res=false;
        }
        return res;
    }

    public boolean checkTriple (Triple test, Triple restriction) {
        boolean res= true;
        if(!test.getRelation().equals(restriction.getRelation())) return false;
        if(!test.getLeft().equals(restriction.getLeft()) && !isRelated(test.getLeft(), restriction.getLeft())) return false;
        if(!test.getRight().equals(restriction.getRight()) && !isRelated(test.getRight(), restriction.getRight())) return false;
        return res;
    }

    public boolean matchTriple (List<Triple> t, Triple test) {
        boolean res=false;
        for(Triple current: t) {
            if(current.getLeft().equals(test.getLeft()) || current.getLeft().equals(test.getRight())) 
                if(isRelated(current.getRight(),test.getLeft()) || isRelated(current.getRight(), test.getRight())) return true;
            if(current.getRight().equals(test.getLeft()) || current.getRight().equals(test.getRelation())) 
                if(isRelated(current.getLeft(),test.getLeft()) || isRelated(current.getLeft(), test.getRight())) return true;
        }
        return res;
    }

    public boolean checkGuard (List<Triple> t, Triple r) {
        boolean res=false;
        if(r.getRelation().equals("==") ) {
            for(Triple outer: t) {
                if(outer.getLeft().equals(r.getLeft())||outer.getLeft().equals(r.getRight())) {
                    for(Triple inner: t) {
                        if(inner.getRelation().equals(outer.getRelation()) && !inner.equals(outer)) {
                            if(isRelated(outer.getLeft(), inner.getLeft()) && isRelated(outer.getRight(), outer.getRight())) {
                                if(inner.getLeft().equals(r.getLeft()) || inner.getLeft().equals(r.getRight())) return true;
                            }
                        }
                    }
                }
                if(outer.getRight().equals(r.getLeft()) || outer.getRight().equals(r.getRight())) {
                    for(Triple inner: t) {
                        if(inner.getRelation().equals(outer.getRelation()) && !inner.equals(outer)) {
                            if(isRelated(outer.getLeft(), inner.getLeft()) && isRelated(outer.getRight(), outer.getRight())) {
                                if(inner.getRight().equals(r.getLeft()) || inner.getRight().equals(r.getRight())) return true;
                            }
                        }
                    }
                }
            }
        } else if(r.getRelation().equals("\\=")) {
            for(Triple outer: t) {
                if(outer.getLeft().equals(r.getLeft())||outer.getLeft().equals(r.getRight())) {
                    for(Triple inner: t) {
                        if(inner.getRelation().equals(outer.getRelation()) && !inner.equals(outer)) {
                            if(isRelated(outer.getLeft(), inner.getLeft()) && isRelated(outer.getRight(), outer.getRight())) {
                                res=true;
                                if(inner.getLeft().equals(r.getLeft()) || inner.getLeft().equals(r.getRight())) return false;
                            }
                        }
                    }
                }
                if(outer.getRight().equals(r.getLeft())||outer.getRight().equals(r.getRight())) {
                    for(Triple inner: t) {
                        if(inner.getRelation().equals(outer.getRelation()) && !inner.equals(outer)) {
                            if(isRelated(outer.getLeft(), inner.getLeft()) && isRelated(outer.getRight(), outer.getRight())) {
                                res=true;
                                if(inner.getRight().equals(r.getLeft()) || inner.getRight().equals(r.getRight())) return false;
                            }
                        }
                    }
                }
            }
        }else {
            for(Triple current: this.triplos) {
                if(checkTriple(current, r)) {
                    if(matchTriple(t,current)){
                        t.add(current.clone());
                        t.add(r.clone());
                        res=true;
                    }
                }
            }
        }


        return res;
    }

    public boolean holdsRestriction(List<Triple> t, Restriction r) {
        boolean res=false;
        Triple rt=r.getTriple();
        Restriction conjunction= r.getConjunctions();
        Restriction disjunction= r.getDisjunctions();
        boolean negative=r.getNegative();
        if(rt!=null) {
            res=checkGuard(t,rt);
            if(res && conjunction!=null){
                System.out.println(t.toString());
                res=holdsRestriction(t,conjunction);
            }            
            if(!res && disjunction!=null) {
                res=holdsRestriction(t, disjunction);
            }
            System.out.println(res);
        } else {
            if(conjunction!=null) {
                if(negative) {
                    res=!holdsRestriction(t, conjunction);
                } else {
                    res=holdsRestriction(t, conjunction);
                }
            }
            if(!res && disjunction!=null) {
                res=holdsRestriction(t, disjunction);
            }
        }
        return res;
    }

    public boolean testRestrictions () {
        boolean res= true;
        for(Triple t: this.triplos) {
            for(Map.Entry<Triple,Restriction> kv: this.restrictions.entrySet()) {
                if(isRestricted(t, kv.getKey()) && !t.equals(kv.getKey())) {
                    List<Triple> lt = new ArrayList<>();
                    lt.add(t);
                    lt.add(kv.getKey());
                    if(!holdsRestriction(lt,kv.getValue())) {
                        res=false;
                        addSemError("Triple "+t.getLeft()+"="+t.getRelation()+"=>"+t.getRight()+ " does not respect all the restrictions.");
                    }
                }
            }
        }
        return res;
    }

    public Concept getConcept (String s) {
        for(Concept c: this.concepts) {
            if(c.getConcept().equals(s)) return c;
        }
        return null;
    }

    public boolean checkTripleSemError (String left, String relation, String right) {
        boolean error=false;
        if(relation.equals("isa")) {
            if(!this.containsConcept(left)) {
                this.addSemError("ERROR: Subject <"+left+"> of ISA relation is not a concept.");
                error=true;
            }
            if(!this.containsConcept(right)) {
                this.addSemError("ERROR: Subject <"+right+"> of ISA relation is not a concept.");
                error=true;
            }
            if(!error) {
                //Inferência atributos
                Concept r= this.getConcept(right);
                Concept l= this.getConcept(left);
                for(DataProperty d: r.getProperties()) {
                    if(!l.containsAtrib(d.getPropName())) {
                        l.addDataProperty(d.clone());
                    }
                }
            }
        } else if(relation.equals("iof")) {
            if(!this.containsIndividual(left)){
                this.addSemError("ERROR: Subject <"+left+"> of IOF relation is not an individual.");
                error=true;
            }
            if(!this.containsConcept(right)) {
                this.addSemError("ERROR: Subject <"+right+"> of IOF relation is not a concept.");
                error=true;
            }
        } else if(relation.equals("pof")) {
            if(!(this.containsConcept(left)||this.containsIndividual(left))) {
                this.addSemError("ERROR: Subject <"+left+"> is neither concept nor individual.");
                error=true;
            }
            if(!(this.containsConcept(right)||this.containsIndividual(right))) {
                this.addSemError("ERROR: Subject <"+right+"> is neither concept nor individual.");
                error=true;
            }
            if(!error) {
                if(this.containsConcept(left)!=this.containsConcept(right)) {
                    this.addSemError("ERROR: POF relation can't be used between a concept and an individual. Can't define POF relation between <"+left+"> and <"+right+">.");
                    error=true;
                }
            }
        } else if(relation.equals("synonym")) {
            if(!(this.containsConcept(left)||this.containsRelation(left))) {
                this.addSemError("ERROR: Subject <"+left+"> is neither concept nor relation.");
                error=true;
            }
            if(!(this.containsConcept(right)||this.containsRelation(right))) {
                this.addSemError("ERROR: Subject <"+right+"> is neither concept nor relation.");
                error=true;
            }
            if(!error) {
                if(this.containsConcept(left)!=this.containsConcept(right)) {
                    this.addSemError("ERROR: Synonym relation can't be used between a concept and a relation. Can't define synonym relation between <"+left+"> and <"+right+">.");
                    error=true;
                }
            }
        } else {
            if(!(this.containsIndividual(left)||this.containsConcept(left))) {
                this.addSemError("ERROR: Subject <"+left+"> is neither concept nor individual.");
                error=true;
            }
            if(!(this.containsIndividual(right)||this.containsConcept(right))) {
                this.addSemError("ERROR: Subject <"+right+"> is neither concept nor individual.");
                error=true;
            }
            if(!this.containsRelation(relation)) {
                this.addSemError("ERROR: Predicate <"+relation+"> is not a defined relation.");
                error=true;
            }
        }
        return error;
    }

    public void createTriple (String left, String relation, String right) {
        boolean error=this.checkTripleSemError(left, relation, right);
        if(!error) {
            Triple t = new Triple(left,relation,right,this);
            boolean repetido=false;
            for(Triple trl: this.triplos) {
                if(t.equals(trl)) repetido=true;
            }
            if(!repetido) this.triplos.add(t);
            else this.addSemError("ERROR: Repeated triple <"+t.getLeft()+","+t.getRelation()+","+t.getRight()+">.");
        }
    }

    public int getConceptNAtr (String s) {
        for(Concept c: this.concepts) {
            if(c.getConcept().equals(s)) return c.getAtrLength();
        }
        return 0;
    }

    public boolean conceptContainsAtrib (String conc, String a) {
        for(Concept c: this.concepts) {
            if(c.getConcept().equals(conc)) {
                return c.containsAtrib(a);
            }
        }
        return false;
    }

    public boolean containsRelation(final String s) {
        boolean res=false;
        for(Relation r: this.relacoes) {if(r.getName().equals(s)) res=true;}
        return res;
    }

    public Relation getRelation (String s) {
        for(Relation r: this.relacoes) {if(r.getName().equals(s)) return r;}
        return null;
    }

    public void addConcept(final String concept) {
        final Concept c = new Concept(concept);
        this.concepts.add(c);
    }

    public void addConcept(Concept c) {
        if(!this.containsConcept(c.getConcept())) {this.concepts.add(c);}
        else {this.erros.add(new SemError("ERROR: Multiple definition of concept <"+ c.getConcept() + ">"));}
    }

    public void addSemError(final String msg) {
        SemError e = new SemError(msg);
        this.erros.add(e);
    }

    public String individuos2OWL () {
        StringBuilder s= new StringBuilder();
        for(String i: this.individuos) {
            s.append("\n\t<Declaration>");
            s.append("\n\t\t<NamedIndividual IRI=\"#");
            s.append(i);
            s.append("\"/>");
            s.append("\n\t</Declaration>\n");
        }
        return s.toString();
    }

    public String relacoes2OWL() {
        StringBuilder s= new StringBuilder();
        for(Relation r:this.relacoes) {
            s.append("\n\t<Declaration>");
            s.append("\n\t\t<ObjectProperty IRI=\"#");
            s.append(r.getName());
            s.append("\"/>");
            s.append("\n\t</Declaration>\n");
            if(r.isInverse()) {
                s.append("\n\t<InverseObjectProperty>\n");
                s.append("\n\t\t<ObjectProperty IRI=\"#");
                s.append(r.getName());
                s.append("\"/>");
                s.append("\"/>");
                s.append("\n\t\t<ObjectProperty IRI=\"#");
                s.append(r.getInverse());
                s.append("\"/>");
                s.append("\n\t</InverseObjectProperty>\n");
            }
            if(r.getDomain()!=null) {
                s.append("\n\t<ObjectPropertyDomain>\n");
                s.append("\n\t\t<ObjectProperty IRI=\"#");
                s.append(r.getName());
                s.append("\"/>");
                s.append("\n\t\t<Class IRI=\"#");
                s.append(r.getDomain());
                s.append("\"/>");
                s.append("\n\t</ObjectPropertyDomain>\n");
            }
            if(r.getCodomain()!=null) {
                s.append("\n\t<ObjectPropertyRange>\n");
                s.append("\n\t\t<ObjectProperty IRI=\"#");
                s.append(r.getName());
                s.append("\"/>");
                s.append("\n\t\t<Class IRI=\"#");
                s.append(r.getCodomain());
                s.append("\"/>");
                s.append("\n\t</ObjectPropertyRange>\n");
            }
        }
        return s.toString();
    }

    public String odlp2OWL() {

        StringBuilder s=new StringBuilder();
        String base="\"http://www.semanticweb.org/"+System.getProperty("user.name")+"/ontologies/2020/10/"+this.name+"\"";
        if(this.erros.size()==0) {
            s.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n\n");
            s.append("<Ontology xmlns=\"http://www.w3.org/2002/07/owl#\"\n");
            s.append("\txml:base=");
            s.append(base);
            s.append("\n");
            s.append("\txmlns:rdf=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"\n");
            s.append("\txmlns:xml=\"http://www.w3.org/XML/1998/namespace\"\n");
            s.append("\txmlns:xsd=\"http://www.w3.org/2001/XMLSchema#\"\n");
            s.append("\txmlns:rdfs=\"http://www.w3.org/2000/01/rdf-schema#\"");
            s.append("\tontologyIRI=" + base + ">\n");
            s.append("\t<Prefix name=\"\" IRI="+base+"/>\n");
            s.append("\t<Prefix name=\"owl\" IRI=\"http://www.w3.org/2002/07/owl#\"/>\n");
            s.append("\t<Prefix name=\"rdf\" IRI=\"http://www.w3.org/1999/02/22-rdf-syntax-ns#\"/>\n");
            s.append("\t<Prefix name=\"xml\" IRI=\"http://www.w3.org/XML/1998/namespace\"/>\n");
            s.append("\t<Prefix name=\"xsd\" IRI=\"http://www.w3.org/2001/XMLSchema#\"/>\n");
            s.append("\t<Prefix name=\"rdfs\" IRI=\"http://www.w3.org/2000/01/rdf-schema#\"/>\n");
            s.append(individuos2OWL());
            for(Concept c:this.concepts) {
                s.append(c.odlp2OWL());
            }
            s.append(relacoes2OWL());
            for(Triple t:this.triplos) {
                s.append(t.triple2OWL());
            }
            s.append("</Ontology>\n");
        } else {
            for(SemError se: this.erros) {
                System.out.println(se.toString());
            }
        }
        return s.toString();
    }

    public String individuos2DOT() {
        StringBuilder s= new StringBuilder();
        for(String i : this.individuos) {
            s.append("\"");
            s.append(i);
            s.append("\" [shape=rectangle, style=filled, color=goldenrod]; \n");
        }
        return s.toString();
    }

    public String concepts2DOT() {
        StringBuilder s= new StringBuilder();
        for(Concept c : this.concepts) {
            s.append(c.odlp2Dot());
        }
        return s.toString();
    }

    public String triples2DOT () {
        StringBuilder s= new StringBuilder();
        for(Triple t : this.triplos) {
            s.append(t.odlp2Dot());
        }
        return s.toString();
    }

    public String dataProperty2DOT () {
        StringBuilder s= new StringBuilder();
        for(DataProperty d : this.atributos) {
            s.append(d.odlp2Dot());
        }
        return s.toString();
    }

    public String odlp2DOT() {
        StringBuilder s = new StringBuilder();
        if(this.erros.size()==0) {
            s.append("digraph ");
            s.append(this.name);
            s.append(" {\n");
            s.append(this.individuos2DOT());
            s.append(this.concepts2DOT());
            s.append(this.triples2DOT());
            s.append(this.dataProperty2DOT());
            s.append("}");
        }
        return s.toString();
    }

    public String individuos2Prolog() {
        StringBuilder res= new StringBuilder();
        for(String s: this.individuos) {
            res.append("individuo(");
            res.append(s);
            res.append(").\n");
        }
        return res.toString();
    }

    public void normRelations() {
        String normal;
        for(Relation r: this.relacoes) {
            if(r.isInverse()) {
                normal = r.getInverse();
                for (Relation tmp : this.relacoes) {
                    if (tmp.getName().equals(normal)) {
                        if (!tmp.isInverse()) {
                            tmp.addInverse(r.getName());
                        }
                        if (tmp.getDomain() == null && r.getCodomain() != null) {
                            tmp.setDomain(r.getCodomain());
                        }
                        if (tmp.getDomain() != null && r.getCodomain() == null) {
                            r.setCodomain(tmp.getDomain());
                        }
                        if (tmp.getCodomain() == null && r.getDomain() != null) {
                            tmp.setCodomain(r.getDomain());
                        }
                        if (tmp.getCodomain() != null && r.getDomain() == null) {
                            r.setDomain(tmp.getCodomain());
                        }
                    }
                }
            }
        }
    }

    public String relacoes2Prolog() {
        StringBuilder res= new StringBuilder();
        for(Relation r: this.relacoes) {
            res.append("relacao(");
            res.append(r.getName());
            res.append(").\n");
        }
        return res.toString();
    }


    public String dataProperty2Prolog() {
        StringBuilder s= new StringBuilder();
        Map<String, List<DataProperty>> tmp= this.atributos.stream()
                                                    .collect(Collectors.groupingBy(DataProperty::getIndName));
        for(Map.Entry<String,List<DataProperty>> kv: tmp.entrySet()) {
            Collections.sort(kv.getValue());
            Iterator i =  kv.getValue().iterator();
            DataProperty d = (DataProperty) i.next();
            s.append("propriedades( ");
            s.append(kv.getKey());
            s.append(" , [");
            s.append(d.getValue());
            while (i.hasNext()) {
                d=(DataProperty) i.next();
                s.append(", ");
                s.append(d.getValue());
            }
            s.append("]).\n");
        }
        return s.toString();
    }


    public String triples2Prolog() {
        StringBuilder res= new StringBuilder();
        List<Triple> l= new ArrayList<>(this.triplos);
        Collections.sort(l);
        Map<String,StringBuilder> tmp = new HashMap<String,StringBuilder>();
        StringBuilder s; Relation r;
        for(Triple t: l) {
            if(tmp.containsKey(t.getRelation())) {
                s = tmp.get(t.getRelation());
            } else {
                s = new StringBuilder();
                tmp.put(t.getRelation(), s);
            }
            s.append(t.getRelation());
            s.append("(");
            s.append(t.getLeft());
            s.append(", ");
            s.append(t.getRight());
            s.append(").\n");
            r= getRelation(t.getRelation());
            if(r.isInverse()) {
                if(tmp.containsKey(r.getInverse())) {
                    s = tmp.get(r.getInverse());
                } else {
                    s=new StringBuilder();
                    tmp.put(r.getInverse(),s);
                }
                s.append(r.getInverse());
                s.append("(");
                s.append(t.getRight());
                s.append(", ");
                s.append(t.getLeft());
                s.append(").\n");
            }
        }
        for(Map.Entry<String,StringBuilder> kv: tmp.entrySet()) {
            res.append(kv.getValue().toString());
        }
        return res.toString();
    }


    public String concepts2Prolog() {
        StringBuilder s= new StringBuilder();
        StringBuilder atributos= new StringBuilder();
        for(Concept c: this.concepts) {
            List<DataProperty> l = c.getProperties();
            s.append("conceito(");
            s.append(c.getConcept());
            s.append(").\n");
            if(l.size()>=1) {
                atributos.append("\tatributos(");
                atributos.append(c.getConcept());
                atributos.append(", [");
                Collections.sort(l);
                Iterator i = l.iterator();
                DataProperty d= (DataProperty) i.next();
                atributos.append(d.getPropName());
                atributos.append(":");
                if(d.getType().equals("xsd:string")) { atributos.append("string");}
                else if(d.getType().equals("list")) {
                    Iterator tmp=d.getRestrictions().iterator();
                    atributos.append("[");
                    String v;
                    if(tmp.hasNext()) {
                        v= (String) tmp.next();
                        atributos.append(v);
                    }
                    while (tmp.hasNext()) {
                        v = (String) tmp.next();
                        atributos.append(", ");
                        atributos.append(v);
                    }
                    atributos.append("]");
                } else {
                    atributos.append(d.getType());
                }
                while(i.hasNext()) {
                    d= (DataProperty) i.next();
                    atributos.append(", ");
                    atributos.append(d.getPropName());
                    atributos.append(":");
                    if(d.getType().equals("xsd:string")) { atributos.append("string");}
                    else if(d.getType().equals("list")) {
                        Iterator tmp=d.getRestrictions().iterator();
                        atributos.append("[");
                        String v;
                        if(tmp.hasNext()) {
                            v= (String) tmp.next();
                            atributos.append(v);
                        }
                        while (tmp.hasNext()) {
                            v = (String) tmp.next();
                            atributos.append(", ");
                            atributos.append(v);
                        }
                        atributos.append("]");
                    } else {
                        atributos.append(d.getType());
                    }
                }
                atributos.append("]).\n");
            }
        }
        s.append(atributos.toString());
        return s.toString();
    }

    public String atom2Prolog(String atom) {
        StringBuilder s = new StringBuilder();
        if(this.containsConcept(atom)) {
            s.append("conceito(");
            s.append(atom);
            s.append(")");
        } else if(this.containsIndividual(atom)) {
            s.append("individuo(");
            s.append(atom);
            s.append(")");
        } else {
            s.append(atom);
        }
        return s.toString();
    }

    public String restrictions2Prolog () {
        StringBuilder s = new StringBuilder();
        for(Map.Entry<Triple,Restriction> kv: this.restrictions.entrySet()) {
            s.append("\nrestricao(");
            s.append(kv.getKey().getRelation());
            s.append("(");
            s.append(atom2Prolog(kv.getKey().getLeft()));
            s.append(",");
            s.append(atom2Prolog(kv.getKey().getRight()));
            s.append(")) :- ");
            s.append(kv.getValue().odlp2Prolog(this));
            s.append(".\n");
        }
        return s.toString();
    }

    public String odlp2Prolog() {
        StringBuilder s= new StringBuilder();
        if(this.erros.size()==0) {
            s.append("not(X):-X,!,no.\nnot(X).\n\n ");
            s.append(this.concepts2Prolog());
            s.append(this.relacoes2Prolog());
            s.append(this.individuos2Prolog());
            s.append(this.triples2Prolog());
            s.append(this.dataProperty2Prolog());
            s.append(this.restrictions2Prolog());
            s.append("classeDe(Coisa, Classe):- iof(Coisa, Classe).\n");
            s.append("classeDe(Coisa, Classe):- isa(Coisa, Classe).\n\n");
            s.append("validaISA:- isa(X,Y), conceitos(X,Y), fail.\n");
            s.append("validaISA:- write(\'Fim da Validação!\'), nl.\n\n");
            s.append("conceitos(X,Y):- conceito(X), conceito(Y), !.\n");
            s.append("conceitos(X,Y):- write(\'Um dos termos \'), write(X), write(\' ou \'), write(Y), write(\'Não é um CONCEITO.\'), nl.\n\n");
            s.append("validaIOF :- iof(X,Y), instconc(X,Y), fail.\n");
            s.append("validaIOF :- write(\'Fim da Validação!\'), nl.\n\n");
            s.append("instconc(X,Y) :- instancia(X), conceito(Y), !.\n");
            s.append("instconc(X,Y) :- write(\'Um dos termos \'), write(X), write(\' ou \'), write(Y), write(\'Não é uma INSTANCIA ou um CONCEITO.\'), nl.\n\n");
        }
        return s.toString();
    }

    public Set<String> mergeSet (Set<String> a, Set<String> b) {
        Set<String> res= new HashSet<String>();
        if(a==null) {return b;}
        if(b==null) {return a;}
        for (String current: a) {res.add(current);}
        for (String current: b) {res.add(current);}
        return res;
    }

    public Map<String, List<Relation>> groupRelations () {
        return this.relacoes.stream().filter(r -> r.getDomain() != null).collect(Collectors.groupingBy(Relation::getDomain));
    }

    public boolean extension(String c) {
        for(Triple t: this.triplos) {
            if (t.getLeft().equals(c) & t.getRelation().equals("isa")) return true;
        }
        return false;
    }

    public String extensions(String c) {
        StringBuilder s= new StringBuilder();
        boolean first=true;
        for(Triple t: this.triplos) {
            if (t.getLeft().equals(c) & t.getRelation().equals("isa")) {
                if(first) {s.append(t.getRight());}
                else {s.append(", "); s.append(t.getRight());}
            };
        }
        return s.toString();
    }

    public String alloyFunctions () {
        StringBuilder res = new StringBuilder();
        res.append("fun id[S: univ] : univ -> univ {\n\t");
        res.append("(S -> S) & iden\n");
        res.append("}\n\n");
        res.append("fun img [R : univ -> univ] : univ -> univ {\n\t");
        res.append("(~R) . R\n");
        res.append("}\n\n");
        res.append("fun ker [R : univ -> univ] : univ -> univ {\n\t");
        res.append("R . (~R)\n");
        res.append("}\n\n");
        return res.toString();
    }

    public String relations2Alloy (Relation r) {
        StringBuilder res = new StringBuilder();
        res.append(r.getName());
        res.append(" : ");
        if(r.isSimple() & r.isEntire()) {
            res.append("one ");
        } else if (r.isSimple()) {
            res.append("lone ");
        } else if (r.isEntire()) {
            res.append("some ");
        }
        res.append(r.getCodomain());
        return res.toString();
    }

    public String facts2Alloy (Relation r) {
        StringBuilder facts = new StringBuilder();
        if (r.isInjective() && r.isSurjective()) {
            facts.append("fact {\n\t");
            facts.append("id[");
            facts.append(r.getCodomain());
            facts.append("] in img[");
            facts.append(r.getName());
            facts.append("], //surjective in ");
            facts.append(r.getName());
            facts.append("\n\t");
            facts.append("ker[");
            facts.append(r.getName());
            facts.append("] in iden //injective in ");
            facts.append(r.getName());
            facts.append("\n}\n\n");
        } else if (r.isInjective()) {
            facts.append("fact {\n\t");
            facts.append("ker[");
            facts.append(r.getName());
            facts.append("] in iden //injective in ");
            facts.append(r.getName());
            facts.append("\n}\n\n");
        } else if (r.isSurjective()) {
            facts.append("fact {\n\t");
            facts.append("id[");
            facts.append(r.getCodomain());
            facts.append("] in img[");
            facts.append(r.getName());
            facts.append("] //surjective in ");
            facts.append(r.getName());
            facts.append("\n}\n\n");
        }
        if(r.isInverse()) {
            facts.append("fact {\n\t");
            facts.append("(~");
            facts.append(r.getInverse());
            facts.append(" in ");
            facts.append(r.getName());
            facts.append(") and \n\t(");
            facts.append(r.getName());
            facts.append(" in ~");
            facts.append(r.getInverse());
            facts.append(")\n\t}\n\n");
        }
        return facts.toString();
    }

    public String odlp2Alloy () {
        StringBuilder res= new StringBuilder();
        res.append("model ");
        res.append(this.name);
        res.append("\n");
        res.append(alloyFunctions());
        StringBuilder facts= new StringBuilder();
        Map<String,List<Relation>> rels = groupRelations();
        for(Map.Entry<String,List<Relation>> kv: rels.entrySet()) {
            String c= kv.getKey();
            if(!this.extension(c)) {
                res.append("sig ");
                res.append(c);
                res.append("{");
                boolean first=true;
                for(Relation r: kv.getValue()) {
                    if(first) {
                        res.append(relations2Alloy(r));
                        facts.append(facts2Alloy(r));
                        first=false;
                    } else {
                        res.append(",\n\t");
                        res.append(relations2Alloy(r));
                        facts.append(facts2Alloy(r));
                    }
                }
                res.append("}\n\n");
            } else {
                res.append("sig ");
                res.append(c);
                res.append(" extends ");
                res.append(this.extensions(c));
                res.append("{");
                boolean first=true;
                for(Relation r: kv.getValue()) {
                    if(first) {
                        res.append(relations2Alloy(r));
                        facts.append(facts2Alloy(r));
                        first=false;
                    } else {
                        res.append(",\n\t");
                        res.append(relations2Alloy(r));
                        facts.append(facts2Alloy(r));
                    }
                }
                res.append("}\n\n");
            }
        }
        for(Concept c: this.concepts) {
            if(!rels.containsKey(c.getConcept())) {
                if(!this.extension(c.getConcept())) {
                    res.append("sig ");
                    res.append(c.getConcept());
                    res.append("{}\n\n");
                } else {
                    res.append("sig ");
                    res.append(c.getConcept());
                    res.append(" extends ");
                    res.append(this.extensions(c.getConcept()));
                    res.append("{}\n\n");
                }
            }
        }
        res.append(facts.toString());
        res.append("run{}\n");
        return res.toString();
    }
}
